from datetime import datetime
from typing import Union
from dateutil import parser
from dateutil.parser import ParserError
from pytz import timezone
from os import remove


def universal_datetime_converter(incoming_datetime: Union[str, datetime], incoming_timezone: timezone = timezone("Etc/UTC")) -> str:
    """
    Converts datetime values of all kinds into a normalized format
    Assumes midnight if datetime does not include time
    """
    universal_format = '%Y-%m-%dT%H:%M:%SZ'

    try:
        datetime_to_convert = parser.parse(str(incoming_datetime))
    except (ParserError, OverflowError):
        try:
            return datetime.utcfromtimestamp(int(incoming_datetime)).strftime(universal_format)
        except Exception:
            return ""

    if datetime_to_convert.tzinfo is None or datetime_to_convert.tzinfo.utcoffset(datetime_to_convert) is None:
        datetime_to_convert = incoming_timezone.normalize(incoming_timezone.localize(datetime_to_convert))

    datetime_to_convert = datetime_to_convert.astimezone(timezone("Etc/UTC"))
    converted_datetime = datetime_to_convert.strftime(universal_format)
    return converted_datetime


def del_kubernetes_healthcheck_file(health_path='/tmp'):
    try:
        remove(f'{health_path}/health')
    except:
        pass


def write_kubernetes_healthcheck_file(health_path='/tmp'):
    del_kubernetes_healthcheck_file(health_path=health_path)
    with open(f'{health_path}/health', 'w') as health_file:
        health_file.write('Healthy')



# from fluvii.exceptions import PartitionsAssigned
# from fluvii.fluvii_app.fluvii_table_app import TableTransaction
# from .nubium_table_app import NubiumTableApp
# import logging
#
#
# LOGGER = logging.getLogger(__name__)
#
#
# class SalesforceRetrieverTransaction(TableTransaction):
#     """
#     Necessary because we aren't usually consuming a message while producing table messages. We can mock out what
#     that message is.
#     """
#     def key(self):
#         return 'timestamp'
#
#     def headers(self):
#         try:
#             return super().headers()
#         except Exception:
#             return {'guid': 'none', 'last_updated_by': 'self'}
#
#     def partition(self):
#         return 0
#
#
# class NubiumSalesforceRetrieverApp(NubiumTableApp):
#     def __init__(self, app_function, consume_topic, produce_topic_schema_dict, app_function_arglist=None):
#         super().__init__(app_function, consume_topic,
#                          produce_topic_schema_dict=produce_topic_schema_dict, app_function_arglist=app_function_arglist,
#                          transaction_cls=SalesforceRetrieverTransaction)
#
#     # def _app_run(self, *args, **kwargs):
#     #     self.producer.poll(0)
#     #     self.transaction._refresh_transaction_fields()
#     #     self.app_function(self.transaction, *self.app_function_arglist)
#     #     self._shutdown = True
#     #     LOGGER.info('stopping event loop...')
#     #     self.transaction.event_loop.stop()
#
#     def _handle_message(self):
#         self.producer.poll(0)
#         self.transaction._init_attrs()
#         self.app_function(self.transaction, *self.app_function_arglist)  # this app doesn't loop, its a one time call
#         self._shutdown = True  #
#         LOGGER.info('stopping event loop...')
#         self.transaction.event_loop.stop()
#
#     def _run(self, **kwargs):
#         unassigned = True
#         while unassigned:
#             try:
#                 self.consume(**kwargs)
#             except NoMessageError:
#                 self._no_message_callback()
#             except PartitionsAssigned:
#                 unassigned = False
#                 self._handle_rebalance()
#         super()._run(**kwargs)
#
#
# # def _app_run(self, *args, **kwargs):
# #     if not self.transaction.message or self.transaction._committed:
# #         self.consume(*args, **kwargs)
# #         self.app_function(self.transaction, *self.app_function_arglist)
# #     self.commit()
# #
# #
# # def _app_run_loop(self, *args, **kwargs):
# #     while not self._shutdown:
# #         try:
# #             LOGGER.debug('Running app function...')
# #             self._app_run(*args, **kwargs)
# #         except NoMessageError:
# #             self.producer.poll(0)
# #             LOGGER.info('No messages!')
# #         except GracefulTransactionFailure:
# #             LOGGER.info("Graceful transaction failure; retrying message with a new transaction...")
# #             self.transaction.abort_active_transaction()
# #         except FatalTransactionFailure:
# #             LOGGER.info("Fatal transaction failure; recreating the producer and retrying message...")
# #             self.transaction.abort_active_transaction()
# #             self.reset_gtfo_producer()


from unittest.mock import call, MagicMock, patch
import os
import socket

from prometheus_client import CollectorRegistry, Gauge
import pytest

from nubium_utils.metrics import MetricsManager, MetricsPusher


class TestMonitor:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.registry = CollectorRegistry()
        self.label_dict = {
            "app": "testing-metrics-app",
            "job": "mock-hostname",
        }
        with patch.dict(os.environ, {
                "NU_HOSTNAME": self.label_dict["job"],
                "NU_APP_NAME": self.label_dict["app"],
                "NU_MP_PROJECT": "testing-metrics",
                "NU_SCHEMA_REGISTRY_URL": "testing-metrics",
        }):
            self.metrics_manager = MetricsManager(
                registry=self.registry,
                metrics_pusher=MagicMock(spec=MetricsPusher),
            )

    def test_gauges_are_created_with_specific_labels(self):
        assert self.metrics_manager._metrics["messages_consumed"].store._labelnames == (
            "app",
            "job",
            "topic",
        )
        assert self.metrics_manager._metrics["message_errors"].store._labelnames == (
            "app",
            "job",
            "exception",
        )
        assert self.metrics_manager._metrics["messages_produced"].store._labelnames == (
            "app",
            "job",
            "topic",
        )
        assert self.metrics_manager._metrics["seconds_behind"].store._labelnames == (
            "app",
            "job"
        )

    def test_utility_functions_set_specific_label_values_on_each_gauge(self):
        self.metrics_manager.set_seconds_behind(11)
        assert self.registry.get_sample_value("seconds_behind", self.label_dict) == 11

    def test_utility_functions_increment_specific_label_values_on_each_gauge(self):
        self.label_dict["topic"] = "test-topic"
        self.metrics_manager.inc_messages_consumed(1, "test-topic")
        assert self.registry.get_sample_value("messages_consumed", self.label_dict) == 1

        self.metrics_manager.inc_messages_produced(1, "test-topic")
        assert self.registry.get_sample_value("messages_produced", self.label_dict) == 1

    def test_increment_message_errors_increases_corresponding_exception_label(self):
        self.metrics_manager.inc_message_errors(ValueError("test-value-error"))
        self.label_dict["exception"] = "ValueError"
        assert self.registry.get_sample_value("message_errors", self.label_dict) == 1

    def test_registered_custom_metrics_are_incrementable(self):
        self.metrics_manager.register_custom_metric("totally_legit", "See! It even has a legit description!")
        self.metrics_manager.inc_custom_metric("totally_legit")
        self.metrics_manager.inc_custom_metric("totally_legit", 7)
        assert self.registry.get_sample_value("totally_legit", self.label_dict) == 8

    def test_unregistered_custom_metrics_raise_error(self):
        with pytest.raises(ValueError) as exc:
            self.metrics_manager.inc_custom_metric("hammer_time")
        assert exc.match(r"hammer_time.*not registered")


class TestMetricsPusher:
    """
    Tests for the class that manages pushing metrics to a metrics cache
    """

    def test_initialization(self):
        """
        Class should initialize with correct arguments
        """
        metrics_pusher = MetricsPusher(
            "test-job", "test-service-name", "test-service-port", "test-pod-port"
        )
        assert metrics_pusher is not None

    def test_set_gateways(self):
        """
        Set gateways returns 2 IP addresses when called with mocked values
        """

        metrics_pusher = MetricsPusher(
            "test-job", "test-service-name", "test-service-port", "test-pod-port"
        )

        # Mock out the socket.getaddrinfo call return value
        addrinfo_mock = [
            (
                socket.AddressFamily.AF_INET,
                socket.SocketKind.SOCK_DGRAM,
                17,
                "",
                ("127.0.0.1", 8080),
            ),
            (
                socket.AddressFamily.AF_INET,
                socket.SocketKind.SOCK_DGRAM,
                6,
                "",
                ("127.0.0.1", 8080),
            ),
            (
                socket.AddressFamily.AF_INET,
                socket.SocketKind.SOCK_DGRAM,
                17,
                "",
                ("127.0.0.2", 8080),
            ),
            (
                socket.AddressFamily.AF_INET,
                socket.SocketKind.SOCK_DGRAM,
                6,
                "",
                ("127.0.0.2", 8080),
            ),
        ]

        with patch("nubium_utils.metrics.metrics_pusher.socket") as socket_patch:
            socket_patch.getaddrinfo.return_value = addrinfo_mock
            metrics_pusher.set_metrics_pod_ips()
            assert metrics_pusher.metrics_pod_ips == {
                "127.0.0.1:test-pod-port",
                "127.0.0.2:test-pod-port",
            }

    def test_push_metrics_called_for_every_gateway(self):
        metrics_pusher = MetricsPusher(
            "test-job", "test-service-name", "test-service-port", "test-pod-port"
        )

        metrics_pusher.metrics_pod_ips = {
            "test-ip-address-1:test-port",
            "test-ip-address-2:test-port",
        }

        with patch(
            "nubium_utils.metrics.metrics_pusher.push_to_gateway"
        ) as push_to_gateway_patch:
            registry = CollectorRegistry()
            metrics_pusher.push_metrics(registry)
            assert (
                call(
                    "test-ip-address-1:test-port",
                    job="test-job",
                    registry=registry,
                    timeout=15,
                )
                in push_to_gateway_patch.call_args_list
            )
            assert (
                call(
                    "test-ip-address-2:test-port",
                    job="test-job",
                    registry=registry,
                    timeout=15,
                )
                in push_to_gateway_patch.call_args_list
            )

    @patch("nubium_utils.metrics.metrics_pusher.socket.getaddrinfo")
    def test_setting_pod_ips_does_not_raise_exceptions(self, mock_getaddrinfo):
        metrics_pusher = MetricsPusher(
            "test-job", "test-service-name", "test-service-port", "test-pod-port"
        )
        metrics_pusher.metrics_pod_ips = {"test-ip-address:test-port"}
        mock_getaddrinfo.side_effect = Exception("mock getaddrinfo failed"),

        metrics_pusher.set_metrics_pod_ips()
